from pycofitness.logging_config.logging_config import LOGGING_CONFIG
from pycofitness.logging_config.logging_config import ConsoleColor as c_color
from pycofitness.mutation import mutation
from pycofitness.mutation import mutation 
from argparse import ArgumentParser
import pathlib
import logging 
import sys 


logger = logging.getLogger(__name__)

class CmdArgs:
    """Defines command line argument variables for plmDCA and pycofitness.
    """
    msa_file = 'msa_file'
    msa_file_help = """Multiple sequence alignment (MSA) file in FASTA format.
    """
    biomolecule = 'biomolecule'
    biomolecule_help = """Type of biomolecule. It should be either protein or RNA 
    in lower or upper case letters.
    """
    verbose_optional = '--verbose'
    verbose_optional_help = """Show logging information on the terminal.
    """
    seqid_optional = '--seqid'
    seqid_optional_help = """Cut-off value of sequences similarity above which they
    are lumped together.
    """
    lambda_h_optional = '--lambda_h'
    lambda_h_optional_help = """Value of fields penalizing constant for L2 
    regularization of fields.
    """
    lambda_J_optional = '--lambda_J'
    lambda_J_optional_help = """Value of couplings penalizing constant for L2 
    regularization of couplings.
    """
    max_iterations_optional = '--max_iterations'
    max_iterations_help = """Number of iterations for gradient decent 
    for negative pseudolikelihood minimization.
    """
    num_threads_optional = '--num_threads'
    num_threads_help = "Number of threads from plmDCA computation"
    output_dir_optional = '--output_dir'
    output_dir_help = """Directory path to which output results are written.
    If the directory is not existing, it will be created. If this path is not provided, an output directory
    is created using the base name of the MSA file, with "output_" prefix
    added to it.
    """     
#End of class CmdArgs

logger = logging.getLogger(__name__)

def configure_logging():
    """Configures logging. When configured, the logging level is INFO and
    messages are logged to stream handler. Log level name are colored whenever
    the terminal supports that. INFO level is Green, WARNING level is Yellow and
    ERROR level is Red.
    """
    logging.config.dictConfig(LOGGING_CONFIG)
    logging.addLevelName(logging.INFO, '{}{}{}'.format(
        c_color.green, logging.getLevelName(logging.INFO), c_color.nocolor))
    logging.addLevelName(logging.WARNING, '{}{}{}'.format(
        c_color.yellow, logging.getLevelName(logging.WARNING), c_color.nocolor))
    logging.addLevelName(logging.ERROR, '{}{}{}'.format(
        c_color.red, logging.getLevelName(logging.ERROR), c_color.nocolor))
    return None

    
def execute_from_command_line(biomolecule, msa_file, 
        seqid = None, lambda_h = None, lambda_J = None, 
        max_iterations = None ,verbose = False, output_dir = None,
        num_threads = None):
    
    if verbose : configure_logging()
    
    mut_inst = mutation.PointMutation(msa_file, biomolecule, seqid=seqid, lambda_h=lambda_h, 
        lambda_J=lambda_J, num_threads=num_threads,
        max_iterations=max_iterations, verbose=verbose
    )
    delphi_dict= mut_inst.delphi_epistatic()
    if output_dir:
        dir_output = pathlib.Path(output_dir)
        dir_output.mkdir()
    else:
        msa_file_base = pathlib.Path(msa_file).stem 
        dir_output = pathlib.Path(f'output_{msa_file_base}')
        dir_output.mkdir()
    outfile = dir_output / ('deltaphi.txt')
    with open(outfile ,'w') as fh:
        fh.write('#site\treference\talternative\tscore\n')
        for i in range(mut_inst.num_sites):
            for res in mut_inst.standard_residues:
                fh.write('{}\t{}\t{}\t{}\n'.format(i+1,mut_inst.refseq.upper()[i], res, delphi_dict[i][res]))
    return None 


def run_mutation():
    """
    """
    parser = ArgumentParser()
    
    parser.add_argument(CmdArgs.biomolecule, help=CmdArgs.biomolecule_help)
    parser.add_argument(CmdArgs.msa_file, help=CmdArgs.msa_file_help)
    parser.add_argument(CmdArgs.seqid_optional, help=CmdArgs.seqid_optional_help, type=float)
    parser.add_argument(CmdArgs.lambda_h_optional, help=CmdArgs.lambda_h_optional_help, type=float)
    parser.add_argument(CmdArgs.lambda_J_optional, help=CmdArgs.lambda_J_optional_help, type=float)
    parser.add_argument(CmdArgs.max_iterations_optional, help=CmdArgs.max_iterations_help, type=int)
    parser.add_argument(CmdArgs.num_threads_optional, help=CmdArgs.num_threads_help, type=int)
    parser.add_argument(CmdArgs.verbose_optional, help=CmdArgs.verbose_optional_help, action='store_true')
    parser.add_argument(CmdArgs.output_dir_optional, help=CmdArgs.output_dir_help)   
    args = parser.parse_args(args=None if sys.argv[1:] else ['--help'])
    args_dict = vars(args)
    
    execute_from_command_line(
        args_dict.get('biomolecule'), 
        args_dict.get('msa_file'), 
        seqid=args_dict.get('seqid'),
        lambda_h=args_dict.get('lambda_h'),
        lambda_J = args_dict.get('lambda_J'),
        max_iterations = args_dict.get('max_iterations'),
        num_threads = args_dict.get('num_threads'),
        output_dir = args_dict.get('output_dir'),
        verbose = args_dict.get('verbose'),
    )

    
    return None 


if __name__ == '__main__':
    run_mutation()





from pycofitness.fasta_reader.fasta_reader import get_alignment_from_fasta_file
import ctypes 
import logging
import os
import glob 
import numpy as np


"""Python wrapper for psuedolikelihood maximization direct coupling analysis (plmDCA).
The gradient decent algorithm is implemented using c++ backend.

Authors: Mehari B. Zerihun, Fabrizio Pucci
"""

logger = logging.getLogger(__name__)


class PlmDCAException(Exception):
    """Implements exceptions related to PlmDCA computation
    """

class PlmDCA:
    """Wraps  the c++ implementation of plmDCA.

    Attributes
    ----------
        plmdca_so : str 
            Path to the plmDCA shared object created from the C++ source code
    """
    plmdca_so_paths  = glob.glob(
        os.path.join(os.path.dirname(__file__), '_plmdcaBackend*')
    )
    logger.info('\n\tplmdca backend path: {}'.format(plmdca_so_paths))
    try:
        plmdca_lib_path = plmdca_so_paths[0]
    except IndexError:
        logger.error('\n\tUnable to find plmdca dynamic library path.' 
            '\nAre you running  pydca as a module before installation?'
            '\nIn this case you need to build the plmdca shared object.'
        )
        raise 


    def __init__(self, msa_file, biomolecule, seqid = None, lambda_h=None, 
            lambda_J = None, max_iterations = None, num_threads = None, 
            verbose = False):
        """Initilizes plmdca instances
        """
        self.__biomolecule = biomolecule.strip().upper()
        if self.__biomolecule not in ('PROTEIN', 'RNA'):
            logger.error('\n\tInvalid biomolecule type {}'.format(self.__biomolecule))
            raise PlmDCAException
        self.__msa_file = msa_file
        self.__biomolecule_int = 1 if self.__biomolecule == 'PROTEIN' else 2
        self.__num_site_states = 21 if self.__biomolecule== 'PROTEIN' else 5
        self.__num_seqs, self.__seqs_len = self._get_num_and_len_of_seqs()
        self.__seqid = 0.8 if seqid is None else seqid 
        if self.__seqid <= 0 or self.__seqid > 1.0: 
            logger.error('\n\t{} is an invalid value of sequences identity (seqid) parameter'.format(self.__seqid))
            raise PlmDCAException 
        self.__lambda_h= 0.2*(self.__seqs_len - 1) if lambda_h is None else lambda_h
        if self.__lambda_h < 0 :
            logger.error('\n\tlambda_h must be a positive number. You passed lambda_h={}'.format(self.__lambda_h))
            raise PlmDCAException  
        self.__lambda_J=  0.2*(self.__seqs_len - 1) if lambda_J is None else lambda_J
        if self.__lambda_J < 0: 
            logger.error('\n\tlambda_J must be a positive number. You passed lambda_J={}'.format(self.__lambda_J))
            raise PlmDCAException
        self.__max_iterations = max_iterations if max_iterations is not None else 100
        self.__num_threads = 1 if num_threads is None else num_threads
        self.__verbose = True if verbose else False  
        # plmdcaBackend interface
        # extern "C" float* plmdcaBackend(unsigned short const biomolecule, unsigned short const num_site_states, 
        # const char* msa_file, unsigned int const seqs_len, float const seqid, float const lambda_h, 
        # float const lambda_J, unsigned int const max_iterations)
        self.__plmdca = ctypes.CDLL(self.plmdca_lib_path)
        self.__plmdcaBackend = self.__plmdca.plmdcaBackend 
        self.__plmdcaBackend.argtypes = (ctypes.c_ushort, ctypes.c_ushort, ctypes.c_char_p, ctypes.c_uint, 
            ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.c_uint, ctypes.c_uint, ctypes.c_bool)
        data_size = (self.__seqs_len * (self.__seqs_len - 1) * (self.__num_site_states ** 2))/2 + self.__seqs_len * self.__num_site_states
        self.__data_size = int(data_size) 
        self.__plmdcaBackend.restype = ctypes.POINTER(ctypes.c_float * self.__data_size)
        #extern "C" void freeFieldsAndCouplings(float*& fields_and_couplings)
        self.freeFieldsAndCouplings = self.__plmdca.freeFieldsAndCouplings
        #self.freeFieldsAndCouplings.argtypes = (ctypes.POINTER(ctypes.c_float),) 
        self.freeFieldsAndCouplings.restype = None 
        log_message="""Created plmDCA instance with:
            biomolecule: {}
            MSA sequence length: {}
            total number of (unweighted) sequences: {}
            sequence identity: {}
            lambda_h: {}
            lambda_J: {}
            gradient decent iterations: {}
            number of threads: {}
        """.format(self.__biomolecule, self.__seqs_len, self.__num_seqs, 
            self.__seqid, self.__lambda_h, self.__lambda_J, self.__max_iterations,
            self.__num_threads,
        )
        logger.info(log_message)
        return None


    @property
    def biomolecule(self):
        """PlmDCA biomolecule attribute 
        """
        return self.__biomolecule


    @property
    def sequence_identity(self):
        """PlmDCA sequence_identity attribute
        """
        return self.__seqid


    @property
    def lambda_h(self):
        """PlmDCA lambda_h attribute
        """
        return self.__lambda_h
    

    @property
    def lambda_J(self):
        """PlmDCA lambda_J attribute
        """
        return self.__lambda_J


    @property
    def max_iterations(self):
        """PlmDCA max_iterations attribute
        """
        return self.__max_iterations


    @property
    def sequences_len(self):
        """PlmDCA sequences_len attribute
        """
        return self.__seqs_len


    @property
    def num_sequences(self):
        """PlmDCA num_sequences attribute
        """
        return self.__num_seqs
    
    
    @property
    def effective_num_sequences(self):
        """PlmDCA effective_num_sequences attribute
        """
        raise NotImplementedError
        

    def _get_num_and_len_of_seqs(self):
        """Obtains the length of sequences in MSA data.

        Parameters
        ----------
            self : PlmDCA 
                Instance of PlmDCA class
        
        Returns
        -------
            num_seqs, seqs_len: tuple  
                A tuple of the number and length of sequences as read from 
                the MSA file
        """
        msa_data = get_alignment_from_fasta_file(self.__msa_file)
        num_seqs = len(msa_data)
        seqs_len = len(msa_data[0])
        return num_seqs, seqs_len


    def map_index_couplings(self, i, j, a, b):
        """Couplings index mapper.

        Parameters
        ----------
            self : PlmDCA 
                An instance of PlmDCA class.
            i : int 
                Site in site-pair (i, j) such that j > i. 
            j : int 
                Site in site-pair (i, j) such that j > i.
        """
        q = self.__num_site_states
        L = self.__seqs_len
        site = int(((L *  (L - 1)/2) - (L - i) * ((L-i)-1)/2  + j  - i - 1) * q * q)
        k =  L * q + site + b + a * q
        return k
    
    
    def get_fields_and_couplings_from_backend(self):
        """Compute fields and couplings using the C++ backend of plmDCA

        Parameters
        ----------
            self : PlmDCA
                An instance of PlmDCA class

        Returns
        -------
            fields_and_couplings : np.array
                A one-dimensional array of the fields and couplings
        """
        logger.info('\n\tComputing fields and couplings using gradient decent')
        h_J_ptr = self.__plmdcaBackend(
            self.__biomolecule_int, self.__num_site_states, self.__msa_file.encode('utf-8'), 
            self.__seqs_len,  self.__seqid, self.__lambda_h, self.__lambda_J, self.__max_iterations,
            self.__num_threads, self.__verbose
        )
        
        fields_and_couplings = np.zeros((self.__data_size,), dtype=np.float32)
        #fields_and_couplings = [c for c in h_J_ptr.contents]
        counter = 0
        for i, h_or_J in enumerate(h_J_ptr.contents):
            fields_and_couplings[i] = h_or_J
            counter += 1
        #Free fields and couplings data from PlmDCABackend
        h_J_ptr_casted = ctypes.cast(h_J_ptr, ctypes.POINTER(ctypes.c_void_p))
        self.freeFieldsAndCouplings(h_J_ptr_casted)

        #fields_and_couplings = np.array(fields_and_couplings, dtype=np.float32)
        
        try:
            assert fields_and_couplings.size == counter 
        except AssertionError:
            logger.error('\n\tData size mismatch from the plmDCA backend')
            raise
        else:
            logger.info('\n\tData size expected from plmDCA backend: {}'.format(self.__data_size))
            logger.info('\n\tData size obtained from plmDCA backend to Python: {}'.format(fields_and_couplings.size))

        return fields_and_couplings 


    def get_couplings_no_gap_state(self, fields_and_couplings_all):
        """Extract the couplings from fields and couplings numpy array

        Parameters
        ----------
            self : PlmDCA 
                An instanc of PlmDCA class
            fields_and_couplings : list/array
                A list of all fields and couplings, including gap state fields
                couplings.
        Returns
        -------
            couplings : np.array
                A one-dimensional array of the couplings
        """
        couplings = list()
        for i in range(self.__seqs_len - 1):
            for j in range(i + 1, self.__seqs_len):
                for a in range(self.__num_site_states - 1):
                    for b in range(self.__num_site_states -1):
                        indx =  self.map_index_couplings(i, j , a, b)
                        couplings.append(fields_and_couplings_all[indx])
        return np.array(couplings)

    
    def get_fields_no_gap_state(self, fields_and_couplings_all):
        """Extracts the fields from fields and couplings numpy array.

        Parameters
        ----------
            self : PlmDCA
                An instance of PlmDCA class
            fields_and_couplings : list/array
                A list of all fields and couplings, including gap state fields
                couplings.
        Returns
        --------
            fields_no_gap_state : list 
                A  list of fields excluding fields corresponding to gap states.
        """
        
        fields_all = fields_and_couplings_all[:self.__seqs_len * self.__num_site_states]
        fields_no_gap_state = list()
        for i in range(self.__seqs_len):
            for a in range(self.__num_site_states - 1): # iterate over q - 1 states to exclude gaps
                fields_no_gap_state.append(fields_all[a + i * self.__num_site_states])
        return fields_no_gap_state

    
    def get_fields_and_couplings_no_gap_state(self, fields_and_couplings_all):
        """Computes fields and couplings excluding gap state fields and gap state
        couplings.

        Parameters
        ----------
            self : PlmDCA 
                An instance of PlmDCA class
            fields_and_couplings : list/array
                A list of all fields and couplings, including gap state fields
                couplings.
        
        Returns
        --------
            fields_no_gap_state, couplings_no_gap_state : tuple
                A tuple of list of fields and couplings
        """

        
        logger.info('\n\tObtaining fields and couplings excluding gap state.')
        fields_no_gap_state = self.get_fields_no_gap_state(fields_and_couplings_all)
        couplings_no_gap_state = self.get_couplings_no_gap_state(fields_and_couplings_all)
        
        return fields_no_gap_state, couplings_no_gap_state

    
    def shift_couplings(self, couplings_ij):
        """Shifts the couplings value.

        Parameters
        ----------
            self : PlmDCA 
                An instance of PlmDCA class
            couplings_ij : np.array
                1d array of couplings for site pair (i, j)
        Returns
        -------
            shifted_couplings_ij : np.array
                A 2d array of the couplings for site pair (i, j)
        """
        qm1 = self.__num_site_states - 1
        couplings_ij = np.reshape(couplings_ij, (qm1,qm1))
        avx = np.mean(couplings_ij, axis=1)
        avx = np.reshape(avx, (qm1, 1))
        avy = np.mean(couplings_ij, axis=0)
        avy = np.reshape(avy, (1, qm1))
        av = np.mean(couplings_ij)
        couplings_ij = couplings_ij -  avx - avy + av
        return couplings_ij     


if __name__ == '__main__':
    """
    """
    



from Bio import AlignIO
from pycofitness.resmapping.resmapping import residue_mapping
import logging

"""Removes gaps from MSA data based of the reference sequence. 
"""

logger = logging.getLogger(__name__)


class MSATrimmer:
    """
    """
    def __init__(self, msa_file, biomolecule):
        """
        """
        self.__msa_file = msa_file 
        self.__biomolecule = biomolecule
        self.__alignment_data = self.get_alignment_data()
        return None 
                
    
    def get_alignment_data(self):
        alignment_data = list(AlignIO.read(self.__msa_file, 'fasta'))
        seq_data_list = list()
        for record in alignment_data:
            seqid = record.id
            seq  = record.seq.upper()
            seq_data_list.append((seqid, str(seq)))
        return seq_data_list
    
    def get_trimmed_cols(self):
        """
        """
        mapping = residue_mapping[self.__biomolecule]
        std_residues = list(mapping.keys())
        trimmed_cols = list()
        refseq = self.__alignment_data[0][1]
        for i, res in enumerate(refseq):
            if res in std_residues: trimmed_cols.append(i)
        return trimmed_cols
       
    def get_trimmed_msa_data(self):
        """Removes all non-standard residues
        """
        trimmed_cols = self.get_trimmed_cols()
        trimmed_msa_data = list()
        for seq_id, seq in self.__alignment_data:
            trimmed_seq = list()
            for i in trimmed_cols:
                trimmed_seq.append(seq[i])
            trimmed_seq = ''.join(trimmed_seq)
            trimmed_msa_data.append((seq_id, trimmed_seq))
        return trimmed_msa_data
    

def main():
    """ 
    """
    test_seq_RNA = '../../tests/tests_input/test_trim_RNA.fa'
    msa_trimmer = MSATrimmer(msa_file=test_seq_RNA, biomolecule='RNA')
    msa_trimmer.get_trimmed_msa_data()
    return None 

if __name__ == '__main__':
    """
    """
    main()




"""A sample configuration for logging. This is a suggestion and you are free to
chose your own preferred way of configuring loggers.
"""

LOGGING_CONFIG = {
    'version':1,
    'disable_existing_loggers':False,
    'formatters':{
        'verbose':{
            'format':'%(levelname)s %(asctime)s %(module)s %(funcName)s %(message)s'
        },
        'simple':{
            'format':'%(levelname)s %(message)s'
        },
    },
    'handlers':{
        'console':{
		    'level':'INFO',
			'class':'logging.StreamHandler',
            'formatter':'verbose',
		},
	},
	'loggers':{
		'':{
			'handlers':['console'],
			'level':'DEBUG',
			'propagate':True,
		},
	},

}

class ConsoleColor:
    """Defines colors for logging messages.

    Attributes
    ----------
        nocolor :  str
            Disable coloring of terminal while logging messages.
        green : str
            This color is used for logging level INFO.
        yellow : str
            This color is used for logging level WARNING.
        red : str
            This color is used for logging level ERROR.
    """
    nocolor = '\033[0;0m'
    green = '\033[0;32m'
    yellow = '\033[0;33m'
    red = '\033[0;31m'



from Bio import AlignIO
from pycofitness.fasta_reader.fasta_reader import get_alignment_from_fasta_file
from pycofitness.resmapping.resmapping import residue_mapping
from logging import config
from  pycofitness.plmdca.plmdca import PlmDCA
from pycofitness.msa_trimmer.msa_trimmer import MSATrimmer
import logging 
from pathlib import Path


logger = logging.getLogger(__name__)


class PlmDCAParams:
    """
    """
    def __init__(self, msa_file:str, biomolecule:str) -> None:
        self.__msa_file = msa_file
        self.__biomolecule = biomolecule
        self.__refseq = get_alignment_from_fasta_file(self.__msa_file)[0]
        return None 
    
    @property
    def refseq(self):
        """
        """
        return self.__refseq
    
    
    def plmdca_params(self, seqid:float=None, lambda_h:float=None, lambda_J:float=None,
            max_iterations:int=None, num_threads:int=None, verbose:bool =False)->tuple:
        """
        """
        logger.info('\n\tComputing fields and couplings using plmDCA algorithm')
        plmDCA_inst =  PlmDCA(self.__msa_file, self.__biomolecule, seqid=seqid, 
            lambda_h=lambda_h, lambda_J=lambda_J, max_iterations=max_iterations,
            num_threads=num_threads, verbose=verbose,
        )
        
        
        fields_and_couplings_all = plmDCA_inst.get_fields_and_couplings_from_backend()
        return plmDCA_inst.get_fields_and_couplings_no_gap_state(fields_and_couplings_all)
    

class PointMutation:
    """
    """
    def __init__(self, msa_file, biomolecule, seqid=None, lambda_h=None, lambda_J=None, num_threads=None,
                 max_iterations=None, verbose=False):
        """ 
        """
        self.__biomolecule = biomolecule.strip().upper() 
        self.__msa_file = self.get_trimmed_msa_file(msa_file, self.__biomolecule)
        self.__seqid = seqid
        self.__lambda_h = lambda_h
        self.__lambda_J = lambda_J
        self.__num_threads = num_threads
        self.__max_iterations = max_iterations
        self.__verbose = verbose
        self.__qm1 = 4 if self.__biomolecule == 'RNA' else 20
        
        return None 
    @property
    def num_sites(self):
        return self.__num_sites
    
    @property
    def refseq(self):
        return self.__refseq 
    
    @property
    def standard_residues(self):
        """ 
        """
        return list(residue_mapping[self.__biomolecule].keys())
    
    def get_trimmed_msa_file(self, orig_msa_file, biomolecule):
        """ 
        """
        alignmed_data = MSATrimmer(msa_file=orig_msa_file, biomolecule=biomolecule).get_trimmed_msa_data()
        trimed_msa_file =  'trimmed_' + Path(orig_msa_file).name
        with open(trimed_msa_file, 'w') as fh:
            for sid, seq in alignmed_data:
                fh.write('>{}\n{}\n'.format(sid, seq))
        return trimed_msa_file
    
    def initialize_fields_and_couplings(self):
        """
        """
        __params_inst= PlmDCAParams(self.__msa_file, self.__biomolecule)
        self.__fields, self.__couplings = __params_inst.plmdca_params(
            seqid=self.__seqid, lambda_h=self.__lambda_h, lambda_J=self.__lambda_J,
            max_iterations=self.__max_iterations,num_threads=self.__num_threads, verbose=self.__verbose
        )
        self.__refseq = __params_inst.refseq
        self.__num_sites = len(self.__refseq)
        return None 
    

    def delphi_epistatic(self):
        """change in energy function is computed from fields and couplings
        """ 
        self.initialize_fields_and_couplings()
        logger.info('\n\tPerforming epistatic model point mutation for sequence\n\t{}'.format(self.__refseq))
        res_mapping =  residue_mapping[self.__biomolecule]
        list_of_standard_residues = list(res_mapping.keys())
        mut_data = dict()
        L = self.__num_sites
        for i in range(self.__num_sites):
            res_background = self.__refseq[i] 
            if res_background not in list_of_standard_residues:
                logger.error(f'\nReference sequence contains non-standard residue at position {i}')
                raise ValueError
            mut_data[i] = dict()
            fields_loc_i = i * self.__qm1
            hi_background = self.__fields[fields_loc_i + res_mapping[res_background]]
            for b in list_of_standard_residues: # carry out mutations at site i
                hi_mutant = self.__fields[fields_loc_i + res_mapping[b]]
                delta_hi_ba = hi_mutant - hi_background
                delta_Jij_total = 0
                for j in range(i): # j < i sites
                    res_at_j = self.__refseq[j]
                    if res_at_j not in list_of_standard_residues: continue
                    couplings_loc_ij = int(((L *  (L - 1)/2) - (L - j) * ((L-j)-1)/2  + i  - j - 1) * self.__qm1 * self.__qm1)
                    k_mutant = couplings_loc_ij + res_mapping[b] + res_mapping[res_at_j] * self.__qm1
                    k_background = couplings_loc_ij + res_mapping[res_background] + res_mapping[res_at_j] * self.__qm1 
                    delta_Jij = self.__couplings[k_mutant] - self.__couplings[k_background]
                    delta_Jij_total += delta_Jij
                for j in range(i + 1, self.__num_sites): # j > i sites
                    res_at_j = self.__refseq[j]
                    if res_at_j not in list_of_standard_residues: continue 
                    couplings_loc_ij = int(((L *  (L - 1)/2) - (L - i) * ((L-i)-1)/2  + j  - i - 1) * self.__qm1 * self.__qm1)
                    k_mutant = couplings_loc_ij + res_mapping[res_at_j]  + res_mapping[b] * self.__qm1
                    k_background = couplings_loc_ij + res_mapping[res_at_j] + res_mapping[res_background] * self.__qm1
                    delta_Jij = self.__couplings[k_mutant] - self.__couplings[k_background]
                    delta_Jij_total += delta_Jij
                deltaPhi = delta_Jij_total + delta_hi_ba
                mut_data[i][b] = deltaPhi
        return mut_data                         

#End of class PointMutation
def test_protein():
    """
    """
    test_msa_Protein = 'TestMSA/TPK1.faclean'
    biomolecule= 'protein'  # or 'rna'
       
    point_mut_inst = PointMutation(test_msa_Protein, biomolecule, max_iterations=20000, num_threads=6, verbose=True)
    delphi_epi=point_mut_inst.delphi_epistatic()
    #delphi_ind = point_mut_inst.delphi_independent()
    counter = 0
    for i in range(len(point_mut_inst.refseq)):
        for a in point_mut_inst.standard_residues:
            counter += 1
            print(counter, delphi_epi[i][a])
    return None

def test_rna():
    test_msa_Protein = 'TestMSA/RF00059.faclean'
    biomolecule= 'rna' 
       
    point_mut_inst = PointMutation(test_msa_Protein, biomolecule, max_iterations=20000, num_threads=6, verbose=True)
    delphi_epi=point_mut_inst.delphi_epistatic()
    #delphi_ind = point_mut_inst.delphi_independent()
    counter = 0
    for i in range(len(point_mut_inst.refseq)):
        for a in point_mut_inst.standard_residues:
            counter += 1
            print(counter, delphi_epi[i][a])
            
    print(point_mut_inst.standard_residues)
    
    return None

def main():
    """ 
    """
    logger.info(f'\n\tYou are running {__file__}')
    test_rna()
    
    
                        
if __name__ == '__main__':
    from pycofitness.logging_config.logging_config import configure_logging
    configure_logging()
    main()



from Bio import AlignIO
import logging 

logger = logging.getLogger(__name__)

def get_alignment_from_fasta_file(file_name):
    """Read sequences from FASTA file using Bio.AlignIO.read()
    Parameters
    ----------
        file_name : str
            Path to FASTA formatted file.
    Returns
    -------
        alignment : list
            A list of biomolecular sequence strings.
    """
    alignment = []
    try:
        record_iterator = AlignIO.read(file_name, 'fasta')
        #biopython just reads the records if there are tags (>some key).
        #It doesn't know if the file is really a biological sequence or not
    except Exception as expt:
        error_msg='\n\tError occured while reading from fasta file: {}.' +\
            '\n\tError type:{}\n\tArguments:{!r}'
        logger.error(error_msg.format(file_name, type(expt).__name__, expt.args))
        raise
    else:
        if any(True for _ in record_iterator):
            for record in record_iterator:
                seq = record.seq.strip()
                if seq: alignment.append(seq.upper())
            if not alignment:
                logger.error(
                    '\n\trecord_iterator returned by AlignIO.read()'
                    ' has no sequences',
                )
                raise ValueError

        else:
            logger.error(
                '\n\trecord_iterator returned by AlignIO.read() is empty',
            )
            raise ValueError
    return alignment


from setuptools import setup, find_packages, Extension
#from setuptools.command.build_ext import build_ext 

with open("README.md") as fh:
    long_description = fh.read()

requirements = [
    'llvmlite>0.30.0',
    'numpy',
    'biopython>=1.74',
]

plmdca_compile_args = ["-fopenmp", "-std=c++11", "-O3"]  
plmdca_link_args = ["-fopenmp", "-O3"] 


plmdca_ext = Extension(
    'pycofitness.plmdca._plmdcaBackend',
    [   
        'pycofitness/plmdca/lbfgs/lib/lbfgs.cpp',
        'pycofitness/plmdca/plmdca_numerics.cpp',
        'pycofitness/plmdca/plmdcaBackend.cpp', 
    ],
    include_dirs=[
        'pycofitness/plmdca/include/',
        'pycofitness/plmdca/lbfgs/include/',
    ],
    extra_compile_args = plmdca_compile_args,
    extra_link_args = plmdca_link_args,
    language = "c++",  
)

setup(
    name="pycofitness",
    version="1.0",
    author="Fabrizio Pucci, Mehari B. Zerihun",
    author_email="Fabrizio.Pucci@ulb.be, mbzerihun@gmail.com",
    python_requires=">=3.6",
    description="<i>In silico</i> mutagenesis of protein and RNA sequences using coevolution",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/KIT-MBS/pycofitness",
    download_url="https://pypi.org/project/pycofitness/",
    packages=find_packages(
        exclude=["*.tests","*.tests.*","tests.*", "tests",
            "examples", "*.examples", "examples.*", "*.examples.*",
        ],
    ),
    ext_modules = [plmdca_ext],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: C++",
        "Programming Language :: C",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",
        "Operating System :: POSIX :: Linux",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    install_requires= requirements,
    tests_require = requirements,
    entry_points={
        "console_scripts":[

            "pycofitness=pycofitness.main:run_mutation",
        ],
    },
    test_suite="tests",
)
